#Author: Joseph Haraldson
#Purpose: Basic implementation of SQP solver routines for constrained newtons method. Solves first order KKT equations...

DEBUG_MODE := true; #for debugging purposes...

#possible methods for solving the SQP are...
#1.) "subset" method - attempt to compute a rank profile of the Jacobian and proceed to proceed as usual in the iteration.
#2.) "trunc" method - attempt to solve the newton step with a truncated SVD algorithm
#3.) "regularize" method - attempt to solve the Newton step using a constraint regularization to obtain a total least norm solution to the Lagrange multiplier problem
#4.) "NAG" method - Use the linear least square solver in maple (NAG libraries)
#5.) "hybrid" method - use a trust region method or something to determine the step behavior.  
#6.)  "vanilla" method - do a plain Newton iteratation without checking assumptions...





singular_poly_sqp_dbg:= proc(A,DA,DA_vars,A_init,K_init,num_iter := 10,tol:=1e-9,divergence:=1e2,method:="vanilla")


	local 	default_DeltaA_pt,
		DA_lin,B_lin,B_ker_lin,
		residuals,
		vars,
		scale_constraints,
		Lambda,
		my_soln,
		Phi,
		cf_vec,cf_mat, #needed for initial guessing
		L,
		B_lin_init,
		DA_init,
		init_pt,
		Z_init,
		Z
		;

	default_DeltaA_pt:= [seq(DA_vars[i]=0,i=1..nops(DA_vars))]:
	DA_lin,B_lin,B_ker_lin:= linearize_A_minimal_kernel(A+DA,t,K_init,'FF'):
	residuals := convert(convert( simplify((DA_lin) .B_lin),set) minus {0},list):
	vars := [ op(DA_vars),op(convert(B_lin,list))]: #same for both of them...
	scale_constraints := [seq( poly_norm_opt(B_ker_lin[1..LinearAlgebra:-RowDimension(B_ker_lin),j],t)^2-1,j=1..LinearAlgebra:-ColumnDimension(B_ker_lin))];    

	residuals := [op(residuals),op(scale_constraints) ];
	Phi := poly_norm_opt(DA,t)^2 + poly_norm_opt(B_ker_lin,t)^2-LinearAlgebra:-ColumnDimension(K_init);
	

	Lambda := [seq('lambda'[i],i=1..nops(residuals))];
	L:=Phi - LinearAlgebra:-DotProduct(convert(Lambda,Vector), convert(residuals,Vector), conjugate=false);

	B_lin_init := compute_initial_guess_kernel(K_init, B_ker_lin,t):

	
	#can't handle this weird linear structure. need to patch.
	DA_init := generate_initial_point(A,A_init,DA,t):
	
        cf_vec,cf_mat := LinearAlgebra:-GenerateMatrix(DA_init, DA_vars);
        
        DA_init := Equate(convert(DA_vars,Vector),Optimization:-LSSolve([cf_mat,cf_vec])[2]);
	

	X_init := [op(B_lin_init),  op(DA_init)];
        print(poly_norm_opt(subs(DA_init, DA),t) );
        
        
        
        
	X:=vars;
        Z:= [op(X),op(Lambda)];
        
        

        
            
            Lambda_init := initial_guess_lagrangian(VectorCalculus:-Gradient(L,X),X_init,Lambda);
                
            
            Z_init := [op(X_init), op(Lambda_init)];
	
	
	if evalb(method = "vanilla") = true then 
	    my_soln := sqp_solve_vanilla(L,X,Lambda,X_init,default_DeltaA_pt,num_iter,divergence,tol);
	elif evalb(method = "sln") = true then 
	    print("BEGINNING SLN SOLVE");
	    my_soln := sqp_solve_sln(L,X,Lambda,X_init,default_DeltaA_pt,num_iter,divergence,tol);
        elif evalb(method = "lm") = true then 
           print("BEGINNING LM SOLVE");
            Z:= [op(X),op(Lambda)];
 
            
            Lambda_init := initial_guess_lagrangian(VectorCalculus:-Gradient(L,X),X_init,Lambda);
                
            
            Z_init := [op(X_init), op(Lambda_init)];
            
            
            my_soln :=  levenberg_marquardt_implicit_merit(VectorCalculus:-Gradient(L,Z),Z,Z_init,tol,num_iter )[2];
            
            
           # my_soln := levenberg_marquardt(poly_norm_opt(VectorCalculus:-Gradient(L,Z),t)^2,VectorCalculus:-Gradient(L,Z),Z,Z_init,tol,num_iter)[2];
	fi: 


		
		
	return [sqrt(subs(my_soln,Phi)), subs(my_soln, DA), subs(my_soln, B_ker_lin),my_soln];
	
	end proc:

#Proc: initial_guess_lagrangian
#Input: stuff we need to compute the Lagrangian (Gradient, initial guess, lambda variables)
#Output: initial guess for Lagrange multipliers (computed via linear least squares approximation)

initial_guess_lagrangian := proc(Gx,X_init,Lambda)
	local lagrangian_system,
	      my_eqns,
	      cf_mat,
	      cf_vec,
	      ineq_cons,
	      ls_soln,
	      Lambda_init
	      ; 

	      

        #solving a linear system for initial guess
	lagrangian_system := subs(convert(subs(X_init,Gx),Vector));
	my_eqns := Equate(lagrangian_system, Vector(ArrayNumElems(lagrangian_system)));
	cf_mat,cf_vec := LinearAlgebra:-GenerateMatrix(my_eqns,Lambda);
	

	
	ls_soln := Optimization:-LSSolve([cf_vec,cf_mat])[2];
	Lambda_init := Equate(convert(Lambda,Vector), ls_soln);

	
	return Lambda_init;
    

end proc:


#Proc: SQP_solve
#input: Hessian at point z, Gradient at point z, list (Vector) of variables Z
#output: Delta Z
#procedure to compute a "Newton Step" or linear system solve for an SQP based method for the rank deficient Jacobian case we encounter.

#for now this is essentially a least squares solution to the problem. 
#1.) Implement and test truncated SVD
#2.) Implement and test rank profile method  - this is implemented in the numeric poly_linalg package.
#3.) 


sqp_solve_vanilla:= proc(L,X,Lambda,X_init,default_DeltaA_pt,num_iter,divergence,tol)

    local 
	  G,
	  H,
	  Gx,
	  Gy,
	  J,
	  ZZ,
	  Z,
	  ZZ_next,
	  Lambda_init,
	  DeltaZ,
	  ZZ_prev,
	  i,
	  Phi
	  ;

	  print("SQP_SUBSET IS RUNNING!");
	  
      #this is the SQP solve routine that uses a subset of the rows of the Jacobian.
    Z:= [op(X),op(Lambda)];
    

    G:= VectorCalculus:-Gradient(L,Z);
    H:=VectorCalculus:-Hessian(L,Z);
    Gx := (VectorCalculus:-Gradient(L,X));
    Gy := (VectorCalculus:-Gradient(L,Lambda));
     J:= VectorCalculus:-Jacobian(Gy,X);
  
    Phi := subs( [seq(Lambda[i]=0,i=1..nops(Lambda))],L); #pretty easy to recover objective function
  
    Lambda_init := initial_guess_lagrangian(Gx,X_init,Lambda); #solve a linear least squares problem here
    
    
    
    Z_init := [op(X_init),op(Lambda_init)];

      
     ZZ:= subs(default_DeltaA_pt, convert(subs(Z_init,Z),Vector));
     ZZ_next := Equate(Z,ZZ);
     
	     for i from 1 to num_iter do
	     
		#solve it via iterative least squares for stability purposes
		#DeltaZ:= Optimization:-LSSolve([subs(ZZ_next,-1*G)  , subs(ZZ_next, H)])[2]; #replace this line with the SQP iteration step.
		DeltaZ:= LinearAlgebra:-LinearSolve(subs(ZZ_next,H),subs(ZZ_next,-1*G));
		
		#print("ls residual is ", Optimization:-LSSolve([subs(ZZ_next,-1*G)  , subs(ZZ_next, H)])[1]);
		
		ZZ_prev := ZZ;
		ZZ:= ZZ+DeltaZ; #update
		ZZ_next := Equate(Z,ZZ);

		mu_0 := norm( subs(ZZ_next,<Gx,Gy>),1); 
		
		
		printf("||X-X_next|| = %e, ||Phi - Phi_next|| = %e, mu= %e\n", norm(ZZ-ZZ_prev,2),abs(poly_norm_opt( subs(ZZ_next ,DA),t) - poly_norm_opt(subs(Equate(Z,ZZ_prev),DA),t)),mu_0);
		
	#	printf("||Lambda||_inf = %e, ||Lambda|| = %e, Lambda_min = %e\n",norm(subs(ZZ_next,convert(Lambda,Vector)),infinity),norm(subs(ZZ_next,convert(Lambda,Vector)),2), min(subs(ZZ_next,Lambda)));
		
		
		if norm(ZZ-ZZ_prev,2) < tol or norm( subs(ZZ_next,<Gx,Gy>),2) < tol or abs(poly_norm_opt( subs(ZZ_next ,DA),t) - poly_norm_opt(subs(Equate(Z,ZZ_prev),DA),t)) < tol  then break fi:
		
		#divergence works for normalized examples...
		if norm(ZZ-ZZ_prev,2) > divergence then WARNING("Newton Diverged"); break; fi:

	     od:
	      printf("Jacobian smallest SV is %e\n", LinearAlgebra:-SingularValues(subs(ZZ_next,J)) [LinearAlgebra:-RowDimension(J)]);
	     
	     return ZZ_next;
      
end proc:









#Proc: SQP_solve_SLN
#Implement the Algorithm from Wright 2005 that has quadratic convergence with a rank deficient Jacobian matrix.
#note the simmilarity to other penalty-like methods.


sqp_solve_sln:= proc(L,X,Lambda,X_init,default_DeltaA_pt,num_iter,divergence,tol)

    local 
	  G,
	  H,
	  Gx,
	  Gy,
	  J,
	  ZZ,
	  Z,
	  ZZ_next,
	  Lambda_init,
	  DeltaZ,
	  ZZ_prev,
	  i,
	  Phi
	  ;

	  print("SQP_SUBSET IS RUNNING!");
	  
      #this is the SQP solve routine that uses a subset of the rows of the Jacobian.
    Z:= [op(X),op(Lambda)];
    
    
    G:= VectorCalculus:-Gradient(L,Z);
    H:=VectorCalculus:-Hessian(L,Z);
    H_orig := H;
    Gx := (VectorCalculus:-Gradient(L,X));
    Gy := (VectorCalculus:-Gradient(L,Lambda));
    J:= VectorCalculus:-Jacobian(Gy,X);
  
    Phi := subs( [seq(Lambda[i]=0,i=1..nops(Lambda))],L); #pretty easy to recover objective function
  
    Lambda_init := initial_guess_lagrangian(Gx,X_init,Lambda); #solve a linear least squares problem here
    
  #  Lambda := convert(Lambda,Vector);
    
    Z_init := [op(X_init),op(Lambda_init)];
    
    
   # print(H[nops(X)+1..LinearAlgebra:-RowDimension(H), nops(X)+1..LinearAlgebra:-ColumnDimension(H)]);
    H[nops(X)+1..LinearAlgebra:-RowDimension(H), nops(X)+1..LinearAlgebra:-ColumnDimension(H)]:= -1*'mu'*Matrix(LinearAlgebra:-IdentityMatrix(LinearAlgebra:-RowDimension(H) - nops(X)));
   # print(H[nops(X)+1..LinearAlgebra:-RowDimension(H), nops(X)+1..LinearAlgebra:-ColumnDimension(H)]);
    
      mu_0 := norm( subs(Z_init,<Gx,Gy>),1); 
    
      
     ZZ:= subs(default_DeltaA_pt, convert(subs(Z_init,Z),Vector));
     ZZ_next := Equate(Z,ZZ);
    printf("||Lambda||_inf = %e, ||Lambda|| = %e, Lambda_min = %e\n",norm(subs(ZZ_next,convert(Lambda,Vector)),infinity),norm(subs(ZZ_next,convert(Lambda,Vector)),2), min(subs(ZZ_next,Lambda)));
		
	     for i from 1 to num_iter do
	     
		mu_0 := norm( subs(ZZ_next,<Gx,Gy>),1); 
		
		
    
		#solve it via iterative least squares for stability purposes
		#DeltaZ:= Optimization:-LSSolve([subs(ZZ_next,-1*G)  , subs(ZZ_next,subs(mu=mu_0, H))])[2]; #replace this line with the SQP iteration step.
		
		DeltaZ:= LinearAlgebra:-LinearSolve(subs(ZZ_next,subs(mu=mu_0, H) ),subs(ZZ_next,-1*G));
		
		#print("ls residual is ", Optimization:-LSSolve([subs(ZZ_next,-1*G)  , subs(ZZ_next, H)])[1]);
		
		ZZ_prev := ZZ;
		ZZ:= ZZ+DeltaZ; #update
		ZZ_next := Equate(Z,ZZ);

		
		printf("||X-X_next|| = %e, ||Phi - Phi_next|| = %e, mu= %e\n", norm(ZZ-ZZ_prev,2),abs(poly_norm_opt( subs(ZZ_next ,DA),t) - poly_norm_opt(subs(Equate(Z,ZZ_prev),DA),t)),mu_0);
		
		#printf("||Lambda||_inf = %e, ||Lambda|| = %e, Lambda_min = %e\n",norm(subs(ZZ_next,convert(Lambda,Vector)),infinity),norm(subs(ZZ_next,convert(Lambda,Vector)),2), min(subs(ZZ_next,Lambda)));
		
		if norm(ZZ-ZZ_prev,2) < tol or mu_0 < tol or abs(poly_norm_opt( subs(ZZ_next ,DA),t) - poly_norm_opt(subs(Equate(Z,ZZ_prev),DA),t)) < tol then break fi:
		
		#divergence works for normalized examples...
		if norm(ZZ-ZZ_prev,2) > divergence then WARNING("Newton Diverged"); break; fi:

	     od:
	     
	     if evalb(DEBUG_MODE = true) then 
		#  printf("The final original Hessian has condition number %e\n", LinearAlgebra:-ConditionNumber(evalf(subs(ZZ_next,H_orig))));
		#  printf("The final perturbed Hessian has condition number %e\n", LinearAlgebra:-ConditionNumber(evalf(subs(ZZ_next,H))));
		  
		  
		  printf("Jacobian smallest SV is %e\n", LinearAlgebra:-SingularValues(subs(ZZ_next,J)) [LinearAlgebra:-RowDimension(J)]);
	     fi:
	     
	     return ZZ_next;
      
end proc:


#the idea is to use levenberg-marquardt here...


