###Numerical differentiation functions
####Author: Joseph Haraldson
###Purpose: Compute Approximate Gradient/Hessian/Jacobians via central finite difference (2nd order approximations). 

read("nearest_smith_functions.mpl"): #Need utility functions

#Proc: numeric_gradient
#Input: Function f: X -> $\RR$ with argument x
#	point x_init to approximate derivative at 
#	h is the step size 
#Output: Gradient of f at the point x.
#Notes: Assume x and x_init are in Vector form instead of list.
#Uses a 4-th order approximation to the derivative
numeric_gradient := proc(f,x_init,x,h:=sqrt(evalf(Digits))) #possibly use a higher-order accuracy method here.
      local i,
	    Df,
	    x_h
	    ;

	    x_h := Vector(ArrayNumElems(x));
	    Df := Vector(ArrayNumElems(x));
	    
	    
	    for i from 1 to ArrayNumElems(x) do 
		  x_h[i] := h; 
		  if i > 1 then x_h[i-1] :=0 fi: #re-use the same vector to reduce memory allocations 
		  
		  
		  
		  Df[i] :=  1./(12*h) * ( -1*subs(Equate(x,x_init+2*x_h),f) +  8*subs(Equate(x,x_init+x_h),f) - 8*subs(Equate(x,x_init-x_h),f) + subs(Equate(x,x_init-2*x_h),f) );

	    od:
	    
	    return Df; 
	    
	    
end proc:

#Input: Vector valued function f
numeric_jacobian := proc(f,x_init,x,h:=sqrt(evalf(Digits)))
	  local i,
		J,
		f_dim,
		x_dim
		;
		f_dim := ArrayNumElems(f);
		x_dim := ArrayNumElems(x);
		J:= Matrix(x_dim,f_dim); #maybe transpose this. Doesn't matter right now. 
		
		for i from 1 to f_dim do 
		      J[i] :=  numeric_gradient(f[i],x_init,x,h);
		od:
		
		return J;
end proc:


#TODO: Write numeric adjoint computation code here


#Proc: symbolic_numeric_adjoint
#Input: Matrix (polynomial) A
#Output: Adjoint of A computed numerically
#Notes:  Uses a symbolic-numeric algorithm that converts A to rationals and returns a numeric approximation
symbolic_numeric_adjoint := proc (A)
	local adj_A
	      ;
	      return evalf(LinearAlgebra:-Adjoint(convert(A,rational)));
end proc:


#Proc: vectorize_adjoint
#Input: Adjoint matrix Adj(A) with indeterminant $t$
#Output: Vectorized adjoint matrix
#Notes: This is needed because the Adjoint operator lives in a O(n^3d) space, as each degree in $t$ is bounded above in degree by $(n-1)d$
#	Dimension Mis-matching could happen in derivative computation
vectorized_adjoint := proc(A,Adj,t)
      local 
	    adj_vec,
	    n,
	    d,
	    degree_bound
	    ;
	    
	    #adj := convert(symbolic_numeric_adjoint(A),Vector); #get the Adjoint matrix via computation.
	    n := LinearAlgebra:-RowDimension(A); #Assume square matrix for now.
	    d := matrix_degree(A,t);
	    degree_bound := (n-1)*d+1;
	    
	    
	    adj_vec:= `<,>` (seq((  Vector( PolynomialTools:-CoefficientVector(Adj(i),t),degree_bound)), i=1..ArrayNumElems(Adj)));
	   
	    
	    return adj_vec; 
end proc:





quit;  #some testing code is down here...

A:= <t^3+6,t^2-3*t,t-1| t^3-1,t+3,t^3| t-1,t^3-1,2*t^3+2*t-1>;

interface(rtablesize=100);

Adj_A := symbolic_numeric_adjoint(A);
vectorized_adjoint(A,Adj_A,t);



quit;
my_func:= x[1]^2+x[2]^2;
vars := <x[1],x[2]>;

x_init := <1,2>;


#Digits:= 30:


Df_approx := numeric_gradient(my_func,x_init,vars,evalf(sqrt(Digits)));

Df := Vector(subs(Equate(vars,x_init),VectorCalculus:-Gradient(my_func,convert(vars,list))));

norm(Df-Df_approx,2);


my_vector_func := <my_func,my_func + x[1]^3>;

J_approx := numeric_jacobian(my_vector_func,x_init,vars);


J_exact := subs(Equate(vars,x_init), VectorCalculus:-Jacobian(my_vector_func,convert(vars,list)));

norm(J_exact-J_approx,2);